/**************************************************
opengate Mac addr auth program

 module to control cache for MAC address 
  the cache holds the temporary copy of DB to speedup checking.
  Only the allowable MAC addresses are hold.

  As ip address check by database is time consuming procedure,
  the recently checked mac addresses are cached.
  Implemented with HashTable.
  HashTable:
    Key= MAC Address
    Val= userId, extraId and cache time.
    If MAC address is found in table and time is new, skip DB access.
    If time is old, remove the cached item.

Copyright (C) 2012 Opengate Project Team
Written by Yoshiaki Watanabe

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Email: watanaby@is.saga-u.ac.jp
**************************************************/
#include "opengatemd.h"

#define MACCACHESIZE 1000000 /* mac cache size */

/* HashTable to store MacAddress->userId,extraId,Time */
static DB* macHashDb;

/* Cache Timeout(seconds) (read from conf file) */
static int macCacheTimeout;

/* Count of hashed items */
static int macCacheItemCount=0;

/****************************************
add item to Mac cache 
****************************************/
int addMacCacheItem(char* macAddress, char* userId, char* extraId) {

  DBT hashKey;
  DBT hashVal;
  char hashValueStr[BUFFMAXLN];

  /* check address format */
  if(isNull(macAddress)) return FALSE;
  if(!ReFormatMacAddr(macAddress)) return FALSE;

  /* check overflow */
  if(macCacheItemCount>MACCACHESIZE) return FALSE;
  macCacheItemCount++;

  /** setup hash key **/
  /* hash key : string of mac address  */  
  hashKey.data = macAddress;
  hashKey.size = strlen(macAddress) + 1;

  /** setup hash value **/
  /* hash value : string "<unixtime>SPACE<userId>SPACE<extraId>" */
  snprintf(hashValueStr,BUFFMAXLN,"%d %s %s",time(NULL),userId,extraId);
  hashVal.data = hashValueStr;
  hashVal.size = strlen(hashValueStr) + 1;    
  if(macHashDb->put(macHashDb, &hashKey, &hashVal, 0) == -1) {
    err_msg("ERR at %s#%d: fail to put into hash table",__FILE__,__LINE__);
    return FALSE;
  }

  return TRUE;
}

/****************************************
query userid and extraid for mac address
if the entry time is old, return false
****************************************/
int queryMacFromMacCache(char* macAddress, char* userId, char* extraId){

  DBT hashKey;
  DBT hashVal;
  time_t entryTime;
  int ret;

  /* set default */
  entryTime=0;
  userId[0]=extraId[0]='\0';

  /* if null or illegal form, return */
  if(isNull(macAddress)) return FALSE;
  if(!ReFormatMacAddr(macAddress)) return FALSE;

  /***** get hashed item matched to the indicated mac */
  hashKey.data = macAddress;
  hashKey.size = strlen(macAddress) + 1;
  memset(&hashVal, 0, sizeof(DBT));
  ret=macHashDb->get(macHashDb, &hashKey, &hashVal, 0);

  /* get is failed, return false */
  if(ret!=0) return FALSE;

  /* get is successed */
  /* pick up the hash values */
  ret=sscanf(hashVal.data,"%d %s %s",&entryTime,userId,extraId);

  /* entryTime and userId should be obtained, but extraId is not */
  if(ret<=1) return FALSE;

  /* if entry time is older than timeout, return false */
  if( entryTime + macCacheTimeout < time(NULL) ) return FALSE;

  return TRUE;
}

/****************************************
initialize Mac Cache
****************************************/
void initMacCache(void) {

  /* prepare hash table */
  if((macHashDb = dbopen(NULL, O_CREAT | O_RDWR, 0644, DB_HASH, NULL)) == NULL) {
    err_msg("ERR at %s#%d: fail to open mac hash table",__FILE__,__LINE__);
    terminateProg(0);
  }

  /* set counter and timeout parameter */
  macCacheItemCount=0;
  if(isNull(GetConfValue("MacCacheTimeout"))){
    err_msg("ERR at %s#%d: cannot get MacCacheTimeout from conf file",__FILE__,__LINE__);
    macCacheTimeout=0;
  }else{
    macCacheTimeout=atoi(GetConfValue("MacCacheTimeout"));
  }  
}

/****************************************
Memory free for Mac Cache
****************************************/
void freeMacCache(void) {

  macHashDb->close(macHashDb);
}

/****************************************
delete item from Mac cache matched to the mac address
****************************************/
int delMacCacheItem(char* macAddress) {

  DBT hashKey;
  
  /* if null or illegal form, return */
  if(isNull(macAddress)) return FALSE;
  if(!ReFormatMacAddr(macAddress)) return FALSE;

  /* delete the item from Hash Table */
  hashKey.data = macAddress;
  hashKey.size = strlen(macAddress) + 1;
  macHashDb->del(macHashDb, &hashKey, 0);

  /* decrement item count */
  if(macCacheItemCount>=0) macCacheItemCount--;

  return TRUE;
}


/****************************************************
 routines for debugging putput
 ***************************************************/

void InitMacCache(void) {
  if(debug>1) err_msg("DEBUG:=>initmacCache( )");
  initMacCache();
  if(debug>1) err_msg("DEBUG:<=initMacCache( )");
}

void FreeMacCache(void) {
  if(debug>1) err_msg("DEBUG:=>freemacCache()");
  freeMacCache();
  if(debug>1) err_msg("DEBUG:<=freeMacCache()");
}

int AddMacCacheItem(char* macAddress, char* userId, char* extraId) {
  int ret;
  if(debug>1) err_msg("DEBUG:=>addMacCacheItem(%s,%s,%s)", macAddress,userId,extraId);
  ret = addMacCacheItem(macAddress,userId,extraId);
  if(debug>1) err_msg("DEBUG:(%d)<=addMacCacheItem( )",ret);
  return ret;
}

int QueryMacFromMacCache(char* macAddress, char* userId, char* extraId){
  int ret;
  if(debug>1) err_msg("DEBUG:=>queryMacFromMacCache(%s)", macAddress);
  ret = queryMacFromMacCache(macAddress, userId, extraId);
  if(debug>1) err_msg("DEBUG:(%d)<=queryMacFromMacCache(,%s,%s)",ret,userId,extraId);
  return ret;
}

int DelMacCacheItem(char* macAddress) {
  int ret;
  if(debug>1) err_msg("DEBUG:=>delMacCacheItem(%s)", macAddress);
  ret = delMacCacheItem(macAddress);
  if(debug>1) err_msg("DEBUG:(%d)<=delMacCacheItem( )",ret);
  return ret;
}

